package org.expeditee.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.awt.image.VolatileImage;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

import org.expeditee.actions.Misc;
import org.expeditee.items.InteractiveWidget;
import org.expeditee.items.Item;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.items.WidgetEdge;

public class FrameGraphics {
	public static final int MESSAGE_BUFFER_HEIGHT = 100;

	private static final int MESSAGE_LINK_Y_OFFSET = 100;

	private static final int MESSAGE_LINK_X = 50;

	// the graphics used to paint with
	private static Graphics2D _DisplayGraphics;

	// the maximum size that can be used to paint on
	private static Dimension _MaxSize;

	// messages shown in the message window
	public static Text[] Messages = new Text[4];

	// buffer of the message window
	private static VolatileImage _MessageBuffer = null;

	// font used for the messages
	private static Font _MessageFont = Font.decode("Serif-Plain-16");

	// the number of messages currently shown (used for scrolling up)
	private static int _MessageCount = 0;

	// modes
	public static final int MODE_NORMAL = 0;

	public static final int MODE_AUDIENCE = 1;

	public static final int MODE_XRAY = 2;

	public static final Color ERROR_COLOR = Color.red;

	private static int _Mode = MODE_NORMAL;

	// if true, error messages are not shown to the user
	private static boolean _SupressMessages = false;

	public static String MESSAGES_FRAMESET_NAME = "Messages";

	// The link to the message frameset
	public static Item MessageLink = new Text(-2, "@" + MESSAGES_FRAMESET_NAME);

	// creator for creating the message frames
	private static FrameCreator _creator;

	/**
	 * If Audience Mode is on this method will toggle it to be off, or
	 * vice-versa. This results in the Frame being re-parsed and repainted.
	 */
	public static void ToggleAudienceMode() {
		Frame current = DisplayIO.getCurrentFrame();
		if (_Mode == MODE_AUDIENCE)
			_Mode = MODE_NORMAL;
		else {
			_Mode = MODE_AUDIENCE;
			ItemUtils.UpdateConnectedToAnnotations(current.getItems());
			for (Overlay o : current.getOverlaysDeep()) {
				ItemUtils.UpdateConnectedToAnnotations(o.Frame.getItems());
			}
			for (Vector v : current.getVectorsDeep()) {
				ItemUtils.UpdateConnectedToAnnotations(v.Frame.getItems());
			}
		}
		FrameUtils.Parse(current);
		DisplayIO.UpdateTitle();
		setMaxSize(new Dimension(_MaxSize.width, _MessageBuffer.getHeight()
				+ _MaxSize.height));
		Repaint();
	}

	/**
	 * If X-Ray Mode is on this method will toggle it to be off, or vice-versa.
	 * This results in the Frame being re-parsed and repainted.
	 */
	public static void ToggleXRayMode() {
		if (_Mode == MODE_XRAY)
			_Mode = MODE_NORMAL;
		else
			_Mode = MODE_XRAY;

		FrameUtils.Parse(DisplayIO.getCurrentFrame());
		DisplayIO.UpdateTitle();
		Repaint();
	}

	/**
	 * @return True if Audience Mode is currently on, False otherwise.
	 */
	public static boolean isAudienceMode() {
		return _Mode == MODE_AUDIENCE;
	}

	/**
	 * @return True if X-Ray Mode is currently on, False otherwise.
	 */
	public static boolean isXRayMode() {
		return _Mode == MODE_XRAY;
	}

	public static void setMaxSize(Dimension max) {
		if (_MaxSize == null)
			_MaxSize = max;

		// Hide the message buffer if in audience mode
		int newMaxHeight = max.height
				- (isAudienceMode() ? 0 : MESSAGE_BUFFER_HEIGHT);
		if (newMaxHeight > 0)
			_MaxSize.setSize(max.width, newMaxHeight);

		if (DisplayIO.getCurrentFrame() != null) {
			DisplayIO.getCurrentFrame().setBuffer(null);
			DisplayIO.getCurrentFrame().setMaxSize(max);
		}

		if (newMaxHeight > 0) {
			_MessageBuffer = null;

			for (int i = 0; i < Messages.length; i++) {
				if (Messages[i] != null) {
					Messages[i].setOffset(0, -_MaxSize.height);
					Messages[i].setMaxSize(_MaxSize);
				}
			}

			MessageLink.setOffset(0, -_MaxSize.height);
			MessageLink.setMaxSize(_MaxSize);
			MessageLink.setPosition(_MaxSize.width - MESSAGE_LINK_Y_OFFSET,
					MESSAGE_LINK_X);
		}
		// Repaint();
	}

	public static Dimension getMaxSize() {
		return _MaxSize;
	}

	public static Dimension getMaxFrameSize() {
		if (DisplayIO.isTwinFramesOn()) {
			return new Dimension((_MaxSize.width / 2), _MaxSize.height);
		} else
			return _MaxSize;
	}

	/**
	 * Sets the Graphics2D object that should be used for all painting tasks.
	 * Note: Actual painting is done by using g.create() to create temporary
	 * instances that are then disposed of using g.dispose().
	 * 
	 * @param g
	 *            The Graphics2D object to use for all painting
	 */
	public static void setDisplayGraphics(Graphics2D g) {
		_DisplayGraphics = g;
	}

	/*
	 * Displays the given Item on the screen
	 */
	private static void PaintItem(Graphics2D g, Item i) {
		if (i == null || g == null)
			return;

		// do not paint annotation items in audience mode
		if (!isAudienceMode()
				|| (isAudienceMode() && !i.isConnectedToAnnotation() && !i
						.isAnnotation()) || i == FrameUtils.LastEdited) {

			Graphics2D tg = (Graphics2D) g.create();
			i.paint(tg);
			tg.dispose();
		}
	}

	public static void AddAllOverlayItems(List<Item> items, Frame overlay,
			List<Frame> seenOverlays) {
		if (seenOverlays.contains(overlay))
			return;

		seenOverlays.add(overlay);

		for (Overlay o : overlay.getOverlays())
			AddAllOverlayItems(items, o.Frame, seenOverlays);

		items.addAll(overlay.getItems());
	}

	/**
	 * Adds all the scaled vector items for a frame into a list. It uses
	 * recursion to get the items from nested vector frames.
	 * 
	 * @param items
	 *            the list into which the vector items will be placed.
	 * @param vector
	 *            the frame containing vecor items.
	 * @param seenVectors
	 *            the vectors which should be ignored to prevent infinate loops.
	 * @param origin
	 *            start point for this frame or null if it is a top level frame.
	 * @param scale
	 *            the factor by which the item on the vector frame are to be
	 *            scaled.
	 */
	public static void AddAllVectorItems(List<Item> items, Frame vector,
			Collection<Frame> seenVectors, Point origin, Float scale) {

		// Check all the vector items and add the items on the vectors

		if (seenVectors.contains(vector))
			return;
		seenVectors.add(vector);

		int originX = origin == null ? 0 : origin.x;
		int originY = origin == null ? 0 : origin.y;

		for (Vector o : vector.getVectors())
			AddAllVectorItems(items, o.Frame, new HashSet<Frame>(seenVectors),
					new Point(originX + Math.round(o.Origin.x * scale), originY
							+ Math.round(o.Origin.y * scale)), o.Scale * scale);
		// if its the original frame then were done
		if (origin == null) {
			ItemUtils.EnclosedCheck(items);
			return;
		}
		// Put copies of the items shifted to the origin of the VectorTag
		int dx = origin.x;
		int dy = origin.y;
		List<Item> copies = ItemUtils.CopyItems(vector.getItems());
		// FrameMouseActions
		Iterator<Item> iterator = copies.iterator();
		while (iterator.hasNext()) {
			Item i = iterator.next();
			// Dont paint annotation items for @v
			if (i.isAnnotation()) {
				iterator.remove();
				continue;
			}
			i.setLinkMark(false);
			i.setActionMark(false);
			if (!(i instanceof Line)) {
				i.setPosition((int) (i.getX() * scale + dx + 0.5), (int) (i
						.getY()
						* scale + dy + 0.5));
				i.setThickness((int) (i.getThickness() * scale + 0.5));
				i
						.setArrowheadLength((int) (i.getArrowheadLength()
								* scale + 0.5));
			}
			if (i instanceof Text)
				i.setSize((int) (i.getSize() * scale + 0.5));
			items.add(i);
		}
	}

	/**
	 * Recursive function similar to AddAllOverlayItems.
	 * 
	 * @param widgets
	 *            The collection the widgets will be added to
	 * @param overlay
	 *            An "overlay" frame - this intially will be the parent frame
	 * @param seenOverlays
	 *            Used for state in the recursion stack. Pass as an empty
	 *            (non-null) list.
	 */
	public static void AddAllOverlayWidgets(List<InteractiveWidget> widgets,
			Frame overlay, List<Frame> seenOverlays) {
		if (seenOverlays.contains(overlay))
			return;

		seenOverlays.add(overlay);

		for (Overlay o : overlay.getOverlays())
			AddAllOverlayWidgets(widgets, o.Frame, seenOverlays);

		widgets.addAll(overlay.getInteractiveWidgets());
	}

	private static Image Paint(Frame toPaint) {
		return Paint(toPaint, true);
	}

	private static Image Paint(Frame toPaint, boolean isActualFrame) {
		if (toPaint == null)
			return null;

		// the buffer is not valid, so it must be recreated
		if (!toPaint.isBufferValid()) {
			Image buffer = toPaint.getBuffer();
			if (buffer == null) {
				GraphicsEnvironment ge = GraphicsEnvironment
						.getLocalGraphicsEnvironment();
				if (!isActualFrame) {
					buffer = new BufferedImage(_MaxSize.width, _MaxSize.height,
							BufferedImage.TYPE_INT_ARGB);
				} else {
					buffer = ge.getDefaultScreenDevice()
							.getDefaultConfiguration()
							.createCompatibleVolatileImage(_MaxSize.width,
									_MaxSize.height);
				}
				toPaint.setBuffer(buffer);
			}

			Graphics2D bg = (Graphics2D) buffer.getGraphics();

			// Nicer looking lines, but may be too jerky while
			// rubber-banding on older machines
			if (UserSettings.AntiAlias)
				((Graphics2D) bg).setRenderingHint(
						RenderingHints.KEY_ANTIALIASING,
						RenderingHints.VALUE_ANTIALIAS_ON);
			// If we are doing @f etc... then have a clear background if its the
			// default background color
			Color backgroundColor = null;
			// Need to allow transparency for frameImages
			if (!isActualFrame) {
				backgroundColor = toPaint.getBackgroundColor();
				if (backgroundColor == null)
					backgroundColor = Item.TRANSPARENT;
			} else {
				backgroundColor = toPaint.getPaintBackgroundColor();
			}

			bg.setColor(backgroundColor);
			bg.fillRect(0, 0, _MaxSize.width, _MaxSize.height);

			List<Item> paintItems = new LinkedList<Item>();
			List<InteractiveWidget> paintWidgets;

			AddAllVectorItems(paintItems, toPaint, new LinkedList<Frame>(),
					null, 1F);

			if (isActualFrame) {
				AddAllOverlayItems(paintItems, toPaint, new LinkedList<Frame>());

				paintWidgets = new LinkedList<InteractiveWidget>();
				AddAllOverlayWidgets(paintWidgets, toPaint,
						new LinkedList<Frame>());

			} else {
				paintItems.addAll(toPaint.getItems());
				paintWidgets = toPaint.getInteractiveWidgets();
			}

			// FIRST: Paint widgets swing gui (not expeditee gui) .
			// Note that these are the ancored widgets
			for (InteractiveWidget iw : paintWidgets) {
				iw.paint(bg);
			}

			PaintPictures(bg, paintItems);

			if (toPaint == DisplayIO.getCurrentFrame())
				PaintPictures(bg, Frame.FreeItems);

			PaintLines(bg, paintItems);
			PaintNonLinesNonPicture(bg, paintItems);

			// toPaint.setBufferValid(true);

			if (DisplayIO.isTwinFramesOn()) {
				List<Item> lines = new LinkedList<Item>();
				for (Item i : Frame.FreeItems) {
					if (i instanceof Line) {
						Line line = (Line) i;

						if (toPaint != DisplayIO.getCurrentFrame()) {
							if (line.getEndItem().isFloating()
									^ line.getStartItem().isFloating()) {
								lines.add(TransposeLine(line,
										line.getEndItem(), toPaint, DisplayIO
												.getRealMouseX(), -DisplayIO
												.getMiddle()));
								lines.add(TransposeLine(line, line
										.getStartItem(), toPaint, DisplayIO
										.getRealMouseX(), -DisplayIO
										.getMiddle()));
							}
						} else {
							if (line.getEndItem().isFloating()
									^ line.getStartItem().isFloating()) {
								Line l = TransposeLine(line, line.getEndItem(),
										toPaint, 0, 0);
								if (l == null)
									l = TransposeLine(line,
											line.getStartItem(), toPaint, 0, 0);
								if (l == null)
									l = line;
								lines.add(l);
							} else
								lines.add(line);
						}
					}
				}
				PaintLines(bg, lines);
			} else {
				// PaintPictures(bg, Frame.FreeItems);
				// PaintNonLinesNonPicture(bg, Frame.FreeItems);
				PaintLines(bg, Frame.FreeItems);
			}

			if (toPaint == DisplayIO.getCurrentFrame())
				PaintNonLinesNonPicture(bg, Frame.FreeItems);

			if (isActualFrame) {
				PaintItem(bg, toPaint.getNameItem());
			}

			// BROOK: Ensure popups are repainted
			if (Browser._theBrowser != null)
				repaintPopups(Browser._theBrowser.getLayeredPane(), bg);

			bg.dispose();
		}

		return toPaint.getBuffer();
	}

	// creates a new line so that lines are shown correctly when spanning
	// across frames in TwinFrames mode
	private static Line TransposeLine(Line line, Item d, Frame toPaint,
			int base, int adj) {
		Line nl = null;

		if (toPaint != DisplayIO.getCurrentFrame() && d.getParent() == null
				&& line.getOppositeEnd(d).getParent() == toPaint) {
			nl = line.copy();
			if (d == line.getStartItem())
				d = nl.getStartItem();
			else
				d = nl.getEndItem();

			if (DisplayIO.FrameOnSide(toPaint) == 0)
				d.setX(base);
			else
				d.setX(base + adj);

		} else if (toPaint == DisplayIO.getCurrentFrame()
				&& d.getParent() == null
				&& line.getOppositeEnd(d).getParent() != toPaint) {
			nl = line.copy();

			if (d == line.getStartItem())
				d = nl.getEndItem();
			else
				d = nl.getStartItem();

			if (DisplayIO.FrameOnSide(toPaint) == 1)
				d.setX(d.getX() - DisplayIO.getMiddle());
			else
				d.setX(d.getX() + DisplayIO.getMiddle());
		}

		return nl;
	}

	private static void Paint(Image left, Image right, Color background) {
		if (_MessageBuffer == null) {
			GraphicsEnvironment ge = GraphicsEnvironment
					.getLocalGraphicsEnvironment();
			_MessageBuffer = ge.getDefaultScreenDevice()
					.getDefaultConfiguration().createCompatibleVolatileImage(
							_MaxSize.width,
							(isAudienceMode() ? 0 : MESSAGE_BUFFER_HEIGHT));
		}

		paintMessage(_MessageBuffer.createGraphics(), background);
		Graphics g = _DisplayGraphics.create();

		// if TwinFrames mode is on, then clipping etc has to be set
		if (DisplayIO.isTwinFramesOn()) {
			// draw the two lines down the middle of the window
			if (left != null)
				left.getGraphics().drawLine(DisplayIO.getMiddle() - 2, 0,
						DisplayIO.getMiddle() - 2, _MaxSize.height);

			if (right != null)
				right.getGraphics().drawLine(0, 0, 0, _MaxSize.height);

			// set the clipping area
			((Graphics2D) g).setClip(0, 0, DisplayIO.getMiddle() - 1,
					_MaxSize.height);
			g.drawImage(left, 0, 0, Item.DEFAULT_BACKGROUND, null);
			((Graphics2D) g).setClip(null);
			g.drawImage(right, DisplayIO.getMiddle() + 1, 0,
					Item.DEFAULT_BACKGROUND, null);

			// otherwise, just draw whichever side is active
		} else {
			if (DisplayIO.getCurrentSide() == 0)
				g.drawImage(left, 0, 0, Item.DEFAULT_BACKGROUND, null);
			else
				g.drawImage(right, 0, 0, Item.DEFAULT_BACKGROUND, null);
		}
		// Dont display the message area in audience mode
		if (!isAudienceMode()) {
			// draw the message area
			g.drawImage(_MessageBuffer, 0, _MaxSize.height, null);
		}
		g.dispose();

	}

	/**
	 * Paints the message area
	 * 
	 * @param g
	 * @param background
	 */
	private static void paintMessage(Graphics2D g, Color background) {
		((Graphics2D) g).setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g.setColor(background);
		g.fillRect(0, 0, _MaxSize.width, MESSAGE_BUFFER_HEIGHT);
		g.setFont(_MessageFont);
		g.setColor(Color.BLACK);
		g.drawLine(0, 0, _MaxSize.width, 0);
		for (Item t : Messages)
			PaintItem(g, t);
		if (MessageLink.getLink() != null)
			PaintItem(g, MessageLink);
		g.dispose();
	}

	public static void Clear() {
		Graphics g = _DisplayGraphics.create();
		g.setColor(Color.WHITE);
		g.fillRect(0, 0, _MaxSize.width, _MaxSize.height);
		g.dispose();
	}

	/**
	 * Called to refresh the display screen.
	 * 
	 */
	public static void Repaint() {
		Runtime.getRuntime();
		if (_DisplayGraphics == null)
			return;

		if (UserSettings.Threading) {
			if (painter == null) {
				painter = new FrameGraphics().new Repainter();

				painter.setDaemon(true);
				painter.setPriority(Thread.MIN_PRIORITY);

				painter.start();
			} else
				painter.run();
		} else {
			Frame[] toPaint = DisplayIO.getFrames();

			if (toPaint != null) {
				Image left = Paint(toPaint[0]);
				Image right = Paint(toPaint[1]);
				Paint(left, right, Item.DEFAULT_BACKGROUND);
			}
		}
	}

	private static Repainter painter = null;

	private static void PaintNonLinesNonPicture(Graphics2D g, List<Item> toPaint) {
		for (Item i : toPaint)
			if (!(i instanceof Line) && !(i instanceof Picture))
				PaintItem(g, i);
	}

	private static void PaintLines(Graphics2D g, List<Item> toPaint) {
		for (Item i : toPaint)
			if (i instanceof Line)
				PaintItem(g, i);
	}

	private static void PaintPictures(Graphics2D g, List<Item> toPaint) {
		for (Item i : toPaint) {
			if (i instanceof Picture) {
				PaintItem(g, i);
			} else if (i.isLineEnd()
					&& (!isAudienceMode() || !i.isConnectedToAnnotation())) {
				i.paintFill(g);
			}
		}
	}

	/**
	 * Highlights an item on the screen Note: All graphics are handled by the
	 * Item itself.
	 * 
	 * @param i
	 *            The Item to highlight.
	 * @param val
	 *            True if the highlighting is being shown, false if it is being
	 *            erased.
	 * @return the item that was highlighted
	 */
	public static Item Highlight(Item i) {
		if ((i instanceof Line)) {
			// Check if within 20% of the end of the line
			Line l = (Line) i;
			Item toDisconnect = l.getEndPointToDisconnect(
					FrameMouseActions.MouseX, FrameMouseActions.MouseY);

			// Brook: Widget Edges do not have such a context
			if (toDisconnect != null && !(i instanceof WidgetEdge)) {
				Item.SelectedMode newMode = toDisconnect.getSelectedMode();
				if (Frame.itemAttachedToCursor())
					newMode = Item.SelectedMode.Normal;
				// unhighlight all the other dots
				for (Item conn : toDisconnect.getAllConnected()) {
					conn.setSelectedMode(Item.SelectedMode.None);
				}
				l.setSelectedMode(newMode);
				// highlight the dot that will be in disconnect mode
				toDisconnect.setSelectedMode(newMode);
				i = toDisconnect;
			} else {
				Collection<Item> connected = i.getAllConnected();
				for (Item conn : connected) {
					conn.setSelectedMode(Item.SelectedMode.Connected);
				}
			}
		} else {
			i.setSelectedMode(Item.SelectedMode.Normal);
		}
		Repaint();
		return i;
	}

	public static void ChangeSelectionMode(Item item, Item.SelectedMode newMode) {
		if (item == null)
			return;

		for (Item i : item.getAllConnected())
			i.setSelectedMode(newMode);
		Repaint();
	}

	public static void OverwriteMessage(String message) {
		for (int ind = Messages.length - 1; ind >= 0; ind--) {
			if (Messages[ind] != null) {
				Messages[ind].setText(getMessagePrefix(false) + message);
				Repaint();
				return;
			}
		}

		// if we have not returned, then there are no messages yet
		DisplayMessage(message, Color.darkGray);
	}

	/**
	 * Displays the given message in the message area of the Frame, any previous
	 * message is cleared from the screen.
	 * 
	 * @param message
	 *            The message to display to the user in the message area
	 */
	public static void DisplayMessage(String message) {
		DisplayMessageAlways(message);
	}

	public static void DisplayMessageOnce(String message) {
		displayMessage(message, null, null, Color.BLACK, false);
	}

	public static void DisplayMessage(String message, Color textColor) {
		displayMessage(message, null, null, textColor);
		// Misc.Beep();
	}

	public static void DisplayMessage(Text message) {
		displayMessage(message.getFirstLine(), message.getLink(), message
				.getAction(), message.getColor());
		// Misc.Beep();
	}

	public static void DisplayMessageAlways(String message) {
		displayMessage(message, null, null, Color.BLACK);
		// Misc.Beep();
	}

	public static void WarningMessage(String message) {
		displayMessage(message, null, null, Color.MAGENTA);
		// Misc.Beep();
	}

	private static String _lastMessage = null;

	private static void displayMessage(String message, String link,
			List<String> actions, Color color) {
		displayMessage(message, link, actions, color, true);
	}

	private static void displayMessage(String message, String link,
			List<String> actions, Color color, boolean displayAlways) {
		assert (message != null);

		if (_SupressMessages)
			return;

		if (!displayAlways && message.equals(_lastMessage)) {
			Misc.Beep();
			return;
		}
		_lastMessage = message;

		if (_creator == null) {
			_creator = new FrameCreator(MESSAGES_FRAMESET_NAME, true);

			// set up 'Messages' link on the right hand side
			MessageLink.setPosition(_MaxSize.width - MESSAGE_LINK_Y_OFFSET,
					MESSAGE_LINK_X);
			MessageLink.setOffset(0, -_MaxSize.height);
		}

		// if the message slots have not all been used yet
		if (_MessageCount <= Messages.length) {
			int pos = 15;
			// find the next empty slot, and create the new message
			for (int i = 0; i < Messages.length; i++) {
				if (Messages[i] == null) {
					Messages[i] = new Text(-1, getMessagePrefix(true) + message);
					Messages[i].setPosition(20, pos);
					Messages[i].setOffset(0, -_MaxSize.height);
					Messages[i].setMaxSize(_MaxSize);
					Messages[i].setColor(color);
					Messages[i].setLink(link);

					_creator.addItem(Messages[i].copy());
					MessageLink.setLink(_creator.getCurrent());
					Repaint();
					return;
				}

				pos += 25;
			}
		}

		// if we have not returned then all message slots are used
		for (int i = 0; i < Messages.length - 1; i++) {
			Messages[i].setText(Messages[i + 1].getFirstLine());
			Messages[i].setColor(Messages[i + 1].getColor());
			Messages[i].setLink(Messages[i + 1].getLink());
		}

		// show the new message
		Text last = Messages[Messages.length - 1];
		last.setColor(color);
		last.setText(getMessagePrefix(true) + message);
		last.setLink(link);
		last.setActions(actions);

		_creator.addItem(last.copy());
		// update the link to the latest message frame
		MessageLink.setLink(_creator.getCurrent());
		Repaint();

		System.out.println(message);
	}

	private static String getMessagePrefix(boolean incrementCounter) {
		if (incrementCounter)
			_MessageCount++;
		return "@" + _MessageCount + ": ";
	}

	/**
	 * Checks if the error message ends with a frame name after the
	 * frameNameSeparator symbol
	 * 
	 * @param message
	 *            the message to be displayed
	 */
	public static void LinkedErrorMessage(String message) {
		if (_SupressMessages)
			return;
		Misc.Beep();
		String[] tokens = message.split(Text.FRAME_NAME_SEPARATOR);
		String link = null;
		if (tokens.length > 1)
			link = tokens[tokens.length - 1];
		displayMessage(message, link, null, ERROR_COLOR);
	}

	public static void ErrorMessage(String message) {
		if (_SupressMessages)
			return;
		Misc.Beep();
		displayMessage(message, null, null, ERROR_COLOR);
	}

	private class Repainter extends Thread {
		public boolean isPainting = false;

		public void run() {
			// TODO see if there is any other way to deal with this
			if (_MaxSize.width <= 0 || _MaxSize.height <= 0) {
				return;
			}
			isPainting = true;

			Frame[] toPaint = DisplayIO.getFrames();

			if (toPaint != null) {
				Image left = Paint(toPaint[0]);
				Image right = Paint(toPaint[1]);
				Paint(left, right, Item.DEFAULT_BACKGROUND);
			}

			isPainting = false;
		}
	}

	/**
	 * Invalidates the buffered image of the current Frame and forces it to be
	 * repainted on to the screen.
	 */
	public static void ForceRepaint() {
		Frame current = DisplayIO.getCurrentFrame();

		if (current == null)
			return;

		Repaint();
	}

	/**
	 * Repaints the buffer of the given Frame.
	 * 
	 * @param toUpdate
	 *            The Frame whose buffer is to be repainted.
	 */

	public static void UpdateBuffer(Frame toUpdate, boolean paintOverlays) {
		if (toUpdate == null)
			return;

		Image vi = Paint(toUpdate, paintOverlays);
		toUpdate.setBuffer(vi);
	}

	public static void SupressMessages(boolean val) {
		_SupressMessages = val;
	}

	private static void repaintPopups(Container parent, Graphics g) {
		for (Component c : parent.getComponents()) {
			if (c instanceof JPopupMenu && ((JPopupMenu) c).isVisible()) {

				Point p = SwingUtilities.convertPoint(c, c.getLocation(),
						Browser._theBrowser.getContentPane());

				g.translate(p.x, p.y);
				c.paint(g);
				g.translate(-p.x, -p.y);
			} else if (c instanceof Container
					&& c != Browser._theBrowser.getContentPane()) {
				repaintPopups((Container) c, g);
			}
		}
	}
}
