package org.expeditee.items;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameGraphics;

/**
 * Implements a line that is drawn on the screen. A line is represented as two
 * Items that share a common Line ID.
 * 
 * @author jdm18
 * 
 */
public class Line extends Item {

	private static final int AUTO_ARROWHEAD_LENGTH = -1;

	private static final int ARROW_HEAD_TO_LENGTH_RATIO = 8;

	private static final int MINIMUM_ARROW_HEAD_LENGTH = 0;

	public static final Float DEFAULT_THICKNESS = 2f;

	public static final Float MINIMUM_THICKNESS = 1f;

	private Item _start;

	private Item _end;

	// old point locations, used for clearing
	private Point _startOffset = new Point(0, 0);

	private Point _endOffset = new Point(0, 0);

	// indicates which end the arrowhead should be drawn at
	private Polygon _poly = null;

	private float _thickness = -1;

	private boolean _isCircle = false;

	// brush strokes used for painting this line and highlighting
	private Stroke _lineStroke = new BasicStroke(DEFAULT_THICKNESS,
			BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);

	// private Stroke _highlightStroke = new BasicStroke (1f + /*GRAVITY +
	// (GRAVITY * (1 / */(2 *GRAVITY)/*))*/, BasicStroke.CAP_BUTT,
	// BasicStroke.JOIN_ROUND);

	/**
	 * Constructs a new Line with the given start and end point and Item ID.
	 * 
	 * @param start
	 *            The starting Point of this Line.
	 * @param end
	 *            The end Point of this Line.
	 * @param id
	 *            The Item ID of this Line.
	 */
	public Line(Item start, Item end, int id) {
		super();
		setID(id);
		_start = start;
		_end = end;

		// update the two end points to add this line
		_start.addLine(this);
		_end.addLine(this);

		// initialise the line size
		float thick = (_start.getThickness() + _end.getThickness()) / 2.0f;

		if (thick < 0)
			thick = 1;
		/*
		 * if (_start instanceof Dot && _end instanceof Dot) thick = (((Dot)
		 * _start).getThickness() + ((Dot) _end) .getThickness()) / 2.0f; else
		 * if (_start instanceof Dot) thick = ((Dot) _start).getThickness();
		 * else if (_end instanceof Dot) thick = ((Dot) _end).getThickness();
		 * else thick = 1;
		 */

		setThickness(thick);
	}

	/**
	 * Returns the Dot at the start of this Line.
	 * 
	 * @return The Dot that is the start of this line.
	 */
	public Item getStartItem() {
		return _start;
	}

	/**
	 * Returns the Dot that is the end of this Line.
	 * 
	 * @return The Dot that is the end of this line.
	 */
	public Item getEndItem() {
		return _end;
	}

	/**
	 * Returns the Dot at the opposite end to the Dot given. If the given Dot is
	 * not part of this line, null is returned.
	 * 
	 * @param fromThis
	 *            The Dot at the opposite end than this Dot is returned
	 * @return The Dot at the opposite end of this Line to the given Dot, or
	 *         Null if the given Dot is not on this Line.
	 */
	public Item getOppositeEnd(Item fromThis) {
		if (_start == fromThis)
			return _end;

		if (_end == fromThis)
			return _start;

		return null;
	}

	/**
	 * Replaces either the start or end point of this Line with the given
	 * LineEnd. This is used when merging LineEnds. It will also add and remove
	 * the line from the LineEnds as well as adjust constraints.
	 * 
	 * @param replace
	 *            Either the start or end Dot of this Line indicating which is
	 *            to be replaced.
	 * @param with
	 *            The Dot to replace the start or end Dot with.
	 */
	public void replaceEnd(Item replace, Item with) {
		Item otherEnd = null;
		if (_start == replace) {
			setStartItem(with);
			otherEnd = _end;
		} else if (_end == replace) {
			setEndItem(with);
			otherEnd = _start;
		}
		// if no end was replaced
		if (otherEnd == null)
			return;
		// Copy the constraints list so the endPoints list can be modified
		List<Constraint> constraints = new ArrayList<Constraint>(replace
				.getConstraints());
		// Now change all the constraints for this line only
		for (int i = 0; i < constraints.size(); i++) {
			Constraint c = constraints.get(i);
			if (c.contains(otherEnd)) {
				c.replaceEnd(replace, with);
			}
		}
	}

	public void setStartItem(Item start) {
		if (start != null) {
			_start.removeLine(this);

			/*
			 * if (!(start instanceof Dot)) { _startOffset.x = _start.getX() -
			 * start.getX(); _startOffset.y = _start.getY() - start.getY(); }
			 * else
			 */{
				_startOffset.x = 0;
				_startOffset.y = 0;
			}

			_start = start;
			_start.addLine(this);
			updatePolygon();
		}
	}

	public void setEndItem(Item end) {
		if (end != null) {
			_end.removeLine(this);

			/*
			 * if (!(end instanceof Dot)) { _endOffset.x = _end.getX() -
			 * end.getX(); _endOffset.y = _end.getY() - end.getY(); } else
			 */{
				_endOffset.x = 0;
				_endOffset.y = 0;
			}

			_end = end;
			_end.addLine(this);
			updatePolygon();
		}
	}

	/**
	 * Returns the Y value of the start Dot of this Line. To determine the end
	 * Dot Y value use getEndDot().getY()
	 * 
	 * @return The Y value of the starting Dot of this Line.
	 * @see #getStartDot()
	 * @see #getEndDot()
	 */
	public int getY() {
		return _start.getY();
	}

	/**
	 * Returns the X value of the start Dot of this Line. To determine the end
	 * Dot X value use getEndDot().getX()
	 * 
	 * @return The X value of the starting Dot of this Line.
	 * @see #getStartDot()
	 * @see #getEndDot()
	 */
	public int getX() {
		// return 0;
		return _start.getX();
	}

	@Override
	public void setX(int x) {
	}

	@Override
	public void setY(int y) {
	}

	@Override
	public void setPosition(int x, int y) {
	}

	@Override
	public String getLink() {
		return null;
	}

	private void paintCircle(Graphics2D g, Item end) {
		int distance = (int) Math.sqrt(Math.pow(Math.abs(_start.getX()
				- _end.getX()), 2)
				+ Math.pow(Math.abs(_start.getY() - _end.getY()), 2));

		// Approximation of circle for mouse interaction
		int points = 20;

		double radians = 0.0;
		int xPoints[] = new int[points];
		int yPoints[] = new int[xPoints.length];

		for (int i = 0; i < xPoints.length; i++) {
			xPoints[i] = (int) (distance * Math.cos(radians));
			yPoints[i] = (int) (distance * Math.sin(radians));
			radians += (2.0 * Math.PI) / xPoints.length;
		}

		_poly = new Polygon(xPoints, yPoints, xPoints.length);
		_poly.translate(end.getX(), end.getY());

		distance *= 2;

		g.drawOval(end.getX() - (distance / 2), end.getY() - (distance / 2),
				distance, distance);
		// Arc version, same result but allows for portions of the circle to be
		// drawn
		// g.drawArc(end.getX() - (distance / 2), end.getY() - (distance / 2),
		// distance, distance, 0, 360);
	}

	public void setLinePattern(int[] pattern) {
		super.setLinePattern(pattern);

		float thick = getThickness();

		if (thick < 0)
			thick = DEFAULT_THICKNESS;

		if (pattern != null) {
			float[] dash = new float[pattern.length];
			for (int i = 0; i < pattern.length; i++)
				dash[i] = (float) pattern[i];
			_lineStroke = new BasicStroke(Math.max(thick, MINIMUM_THICKNESS),
					BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10f, dash,
					0.0f);
		} else
			_lineStroke = new BasicStroke(Math.max(thick, MINIMUM_THICKNESS),
					BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);

		if (_start.getLinePattern() != pattern)
			_start.setLinePattern(pattern);

		if (_end.getLinePattern() != pattern)
			_end.setLinePattern(pattern);
	}

	@Override
	public void paint(Graphics2D g) {
		g.setColor(getPaintColor());

		// if (this._mode == Item.SelectedMode.Disconnect)
		// System.out.println("Disconnect mode!");

		if (_start instanceof Text && ((Text) _start).startsWith("@c")) {
			g.setStroke(_lineStroke);
			paintCircle(g, _end);
			_isCircle = true;
		} else if (_end instanceof Text && ((Text) _end).startsWith("@c")) {
			g.setStroke(_lineStroke);
			paintCircle(g, _start);
			_isCircle = true;
		} else {
			// paint the arrowhead (if necessary)
			g.setStroke(new BasicStroke(getThickness(), BasicStroke.CAP_BUTT,
					BasicStroke.JOIN_MITER));

			paintArrow(g, _start, _startOffset, _endOffset);
			paintArrow(g, _end, _endOffset, _startOffset);

			g.setStroke(_lineStroke);
			g.drawLine(_start.getX() + _startOffset.x, _start.getY()
					+ _startOffset.y, _end.getX() + _endOffset.x, _end.getY()
					+ _endOffset.y);

			_isCircle = false;
		}

		// update the highlight shape
		// java.awt.geom.Line2D.Float shape = new
		// java.awt.geom.Line2D.Float(_startx, _starty, _endx, _endy);
		// _shape = _highlightStroke.createStrokedShape(shape);

		// g.setStroke(norm);

		if (showLineHighlight() && isHighlighted()) {
			g.setColor(getHighlightColor());
			g.setStroke(new BasicStroke(MINIMUM_THICKNESS,
					BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER));
			((Graphics2D) g).draw(this.getArea());
		}

		if (_virtualSpot != null) {
			_virtualSpot.paint(g);
			_virtualSpot = null;
		}
	}

	/**
	 * Based on code from DeSL (Arrow2D) http://sourceforge.net/projects/desl/
	 */
	private void paintArrow(Graphics2D g, Item withArrow, Point startOffset,
			Point endOffset) {
		// only draw an arrowhead if necessary
		if (!(this._mode == Item.SelectedMode.Disconnect && withArrow._mode == Item.SelectedMode.Disconnect)
				&& (!withArrow.hasVisibleArrow() || withArrow.getLines().size() > 1))
			return;

		int x0, x1, y0, y1;

		x1 = withArrow.getX() + startOffset.x;
		y1 = withArrow.getY() + startOffset.y;
		x0 = getOppositeEnd(withArrow).getX() + endOffset.x;
		y0 = getOppositeEnd(withArrow).getY() + endOffset.y;

		int arrowLength = withArrow.getArrowheadLength();
		double arrowRatio = withArrow.getArrowheadRatio();

		// set the size of the disconnect indicator arrowhead
		if (this._mode == Item.SelectedMode.Disconnect) {
			// System.out.println("Disconnect mode!");
			arrowLength = 15;
			arrowRatio = 0.3;
			// Make sure the arrowhead positions get recalculated
			withArrow.setArrowhead(null);
		}

		// if the arrowhead is 'auto', then one and only one end must be
		// floating
		if (arrowLength != AUTO_ARROWHEAD_LENGTH && arrowRatio < 0)
			return;

		int deltaX = x1 - x0;
		int deltaY = y1 - y0;

		int length = (int) Math.sqrt(Math.pow(deltaX, 2) + Math.pow(deltaY, 2));

		if (arrowLength < 0) {
			arrowLength = length / ARROW_HEAD_TO_LENGTH_RATIO;
			arrowLength = Math.max(MINIMUM_ARROW_HEAD_LENGTH, arrowLength);
			withArrow.setArrowhead(null);
			if (arrowRatio < 0)
				arrowRatio = DEFAULT_ARROWHEAD_RATIO;
		}

		// only calculate the arrowhead polygon if necessary
		if (withArrow.getArrowhead() == null) {
			Polygon arrowhead = new Polygon();
			arrowhead.addPoint((int) x0, (int) y0);
			arrowhead.addPoint((int) x0 - arrowLength,
					(int) (y0 - (arrowLength * arrowRatio)));
			arrowhead.addPoint((int) x0, (int) y0);
			arrowhead.addPoint((int) x0 - arrowLength,
					(int) (y0 + (arrowLength * arrowRatio)));
			withArrow.setArrowhead(arrowhead);
		}

		Polygon arrowhead = withArrow.getArrowhead();
		double rad = calcAngle((float) x0, (float) y0, (float) x1, (float) y1);
		arrowhead.translate((x0 + length) - arrowhead.xpoints[0], y0
				- arrowhead.ypoints[0]);
		AffineTransform tx = AffineTransform.getRotateInstance(rad, x0, y0);
		g.draw(tx.createTransformedShape((Shape) arrowhead));
	}

	/**
	 * Calcuates the angle of rotation (in radians) of the arrowhead Based on
	 * code from DeSL (Arrow2D) http://sourceforge.net/projects/desl/
	 */
	private double calcAngle(float x1, float y1, float x2, float y2) {
		double rad = 0.0d;
		float dx = x2 - x1;
		float dy = y2 - y1;

		if (dx == 0.0) {
			if (dy == 0.0) {
				rad = 0.0;
			} else if (dy > 0.0) {
				rad = Math.PI / 2.0;
			} else {
				rad = Math.PI * 3.0 / 2.0;
			}
		} else if (dy == 0.0) {
			if (dx > 0.0) {
				rad = 0.0;
			} else {
				rad = Math.PI;
			}
		} else {
			if (dx < 0.0) {
				rad = Math.atan(dy / dx) + Math.PI;
			} else if (dy < 0.0) {
				rad = Math.atan(dy / dx) + (2 * Math.PI);
			} else {
				rad = Math.atan(dy / dx);
			}
		}

		return rad;
	}

	@Override
	public Polygon getPolygon() {
		if (_poly == null)
			updatePolygon();

		return _poly;
	}

	@Override
	public Line copy() {
		Item s = _start.copy();
		Item e = _end.copy();

		Line temp = new Line(s, e, getID());

		Item.DuplicateItem(this, temp);

		temp.setLinePattern(getLinePattern());
		temp.setThickness(getThickness());

		temp.setArrow(getArrowheadLength(), getArrowheadRatio());

		return temp;
	}

	public Item getEndPointToDisconnect(int x, int y) {
		Item start = getStartItem();
		Item end = getEndItem();
		int deltaX = start.getX() - end.getX();
		int deltaY = start.getY() - end.getY();
		int lineLength = (int) Math.sqrt(deltaX * deltaX + deltaY * deltaY);

		int startX = start.getX() - x;
		int endX = end.getX() - x;
		int startY = start.getY() - y;
		int endY = end.getY() - y;

		int distStart = (int) Math.sqrt(startX * startX + startY * startY);
		int distEnd = (int) Math.sqrt(endX * endX + endY * endY);

		final double DISCONNECT_THRESHHOLD = Math.min(0.2 * lineLength, 25.0);
		final double NORMAL_THRESHHOLD = Math.min(0.1 * lineLength, 13.0);

		if (distStart < NORMAL_THRESHHOLD) {
			start._mode = Item.SelectedMode.Normal;
			return start;
		} else if (distEnd < NORMAL_THRESHHOLD) {
			end._mode = Item.SelectedMode.Normal;
			return end;
		} else if (distStart < DISCONNECT_THRESHHOLD) {
			if (start.getLines().size() > 1
					|| start.getConstraints().size() > 0)
				start._mode = Item.SelectedMode.Disconnect;
			else
				start._mode = Item.SelectedMode.Normal;
			return start;
		} else if (distEnd < DISCONNECT_THRESHHOLD) {
			if (end.getLines().size() > 1 || end.getConstraints().size() > 0)
				end._mode = Item.SelectedMode.Disconnect;
			else
				end._mode = Item.SelectedMode.Normal;
			return end;
		}

		return null;
	}

	@Override
	public int setSelectionColor() {
		super.setSelectionColor();

		return Item.UNCHANGED_CURSOR;
	}

	@Override
	public void setMaxSize(Dimension d) {
	}

	/**
	 * Sets the Color of this Line and the two Dots at the ends.
	 * 
	 * @param c
	 *            The Color to paint this Line in.
	 */
	@Override
	public void setColor(Color c) {
		super.setColor(c);

		_start.lineColorChanged(c);
		_end.lineColorChanged(c);
	}

	@Override
	public void setThickness(float thick) {
		_thickness = thick;

		if (thick < 0)
			thick = DEFAULT_THICKNESS;

		if (_start.getThickness() != thick)
			_start.setThickness(thick);

		if (_end.getThickness() != thick)
			_end.setThickness(thick);

		int[] pattern = _start.getLinePattern();
		if (pattern == null)
			pattern = _end.getLinePattern();

		if (pattern != null) {
			float[] dash = new float[pattern.length];
			for (int i = 0; i < pattern.length; i++)
				dash[i] = (float) pattern[i];
			_lineStroke = new BasicStroke(Math.max(thick, MINIMUM_THICKNESS),
					BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10f, dash,
					0.0f);
		} else
			_lineStroke = new BasicStroke(Math.max(thick, MINIMUM_THICKNESS),
					BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER);
	}

	@Override
	public float getThickness() {
		return _thickness;
	}

	@Override
	public void setAnnotation(boolean val) {
		if (_end.isAnnotation() != val)
			_end.setAnnotation(val);
		else if (_start instanceof Text && _start.isAnnotation() != val)
			_start.setAnnotation(val);
	}

	@Override
	public boolean isAnnotation() {
		return _start.isAnnotation() || _end.isAnnotation();
	}

	@Override
	public boolean isConnectedToAnnotation() {
		return _start.isConnectedToAnnotation()
				|| _end.isConnectedToAnnotation();
	}

	/**
	 * Fixes the length of the arrowhead on the end of the line.
	 */
	public void fixArrowheadLength() {
		if (_end.getArrowheadLength() == AUTO_ARROWHEAD_LENGTH) {
			fixArrowheadLength(_end);
		}
	}

	public void autoArrowheadLength() {
		_end.setArrowheadLength(AUTO_ARROWHEAD_LENGTH);
	}

	public void fixArrowheadLength(Item arrow) {
		// if the arrow isnt there turn it on!
		int x0, x1, y0, y1;

		x1 = arrow.getX();
		y1 = arrow.getY();
		x0 = getOppositeEnd(arrow).getX();
		y0 = getOppositeEnd(arrow).getY();

		int deltaX = x1 - x0;
		int deltaY = y1 - y0;

		int length = (int) Math.sqrt(Math.pow(deltaX, 2) + Math.pow(deltaY, 2));
		length /= ARROW_HEAD_TO_LENGTH_RATIO;
		arrow.setArrowheadLength(Math.max(MINIMUM_ARROW_HEAD_LENGTH, length));
	}

	/**
	 * This method toggles the arrow on the floating end point of a line. If the
	 * line does not have one floating end and one fixed end this method does
	 * nothing.
	 * 
	 */
	public void toggleArrow() {
		Item arrow;
		if (_start.isFloating() && !_end.isFloating())
			arrow = _start;
		else if (_end.isFloating() && !_start.isFloating())
			arrow = _end;
		else
			return;

		// if there already an arrow, turn it off
		if (arrow.getArrowheadLength() != 0 && arrow.getArrowheadRatio() > 0) {
			arrow.setArrowheadLength(0);
			return;
		}

		fixArrowheadLength(arrow);
	}

	private static int[] LinePatterns = new int[] { 0, 10, 20 };

	/**
	 * The rotates through a wheel of dashed lines.
	 * 
	 * @param amount
	 *            number of rotations around the wheel to toggle by.
	 */
	public void toggleDashed(int amount) {
		// find the index of the current line pattern
		int[] currentPattern = getLinePattern();

		// Find the current pattern and move to the next pattern in the wheel
		for (int i = 0; i < LinePatterns.length; i++) {
			if (currentPattern == null || currentPattern[0] == LinePatterns[i]) {
				i += LinePatterns.length + amount;
				i %= LinePatterns.length;

				// if we are at the start of the wheel make it 'null' (solid
				// line)
				if (i == 0) {
					setLinePattern(null);
				} else {
					setLinePattern(new int[] { LinePatterns[i], LinePatterns[i] });
				}
				return;
			}
		}
	}

	private static double[] ArrowHeadRatios = new double[] { 0.25, 0.5, 1.0 };

	/**
	 * This will toggle the arrow head between the various types of UML arrow
	 * heads.
	 * 
	 * @param amount
	 */
	public void toggleArrowHeadRatio(int amount) {
		Item arrow = null;
		if (_start.isFloating() && !_end.isFloating())
			arrow = _start;
		else if (_end.isFloating() && !_start.isFloating())
			arrow = _end;
		else
			return;

		// find the index of the current line pattern
		double currentRatio = arrow.getArrowheadRatio();

		// Find the current pattern and move to the next pattern in the wheel
		for (int i = 0; i < ArrowHeadRatios.length; i++) {
			if (currentRatio - ArrowHeadRatios[i] < 0.001) {
				i += ArrowHeadRatios.length + amount;
				i %= ArrowHeadRatios.length;
				arrow.setArrowheadRatio(ArrowHeadRatios[i]);
				return;
			}
		}
	}

	private static int[] ArrowHeadLength = new int[] { /* AUTO_ARROWHEAD_LENGTH, */
	0, 20, 40, 80 };

	/**
	 * Changes the length of the arrow head. In the future this may toggle the
	 * arrow head between the various types of UML arrow heads.
	 * 
	 * @param amount
	 */
	public void toggleArrowHeadLength(int amount) {
		Item arrow = null;
		if (_start.isFloating() && !_end.isFloating())
			arrow = _start;
		else if (_end.isFloating() && !_start.isFloating())
			arrow = _end;
		else
			return;

		// find the index of the current line pattern
		int currentLength = arrow.getArrowheadLength();

		// Find the current pattern and move to the next pattern in the wheel
		for (int i = 0; i < ArrowHeadLength.length; i++) {
			if (currentLength <= ArrowHeadLength[i]) {
				i += ArrowHeadLength.length + amount;
				i %= ArrowHeadLength.length;
				arrow.setArrowheadLength(ArrowHeadLength[i]);
				return;
			}
		}
	}

	private Item _virtualSpot = null;

	public void showVirtualSpot(Item orig, int mouseX, int mouseY) {
		if (orig.getLines().size() != 1)
			return;

		// lines that are in 'connected' mode, also cannot be attached
		if (orig.getLines().get(0).getOppositeEnd(orig).isFloating())
			return;

		Item spot = new Dot(-1, orig.getX(), orig.getY());
		Item.DuplicateItem(orig, spot);
		spot.setThickness(Math.max(this.getThickness(), 5));
		if (this.getColor() != Color.RED)
			spot.setColor(Color.RED);
		else
			spot.setColor(Color.BLUE);

		// unhighlight all the dots
		for (Item conn : getAllConnected()) {
			conn.setSelectedMode(Item.SelectedMode.None);
		}

		// calculate nearest point on line from spot
		double slope1;
		double slope2;
		double x, y;

		slope1 = (_start.getY() - _end.getY() * 1.0)
				/ (_start.getX() - _end.getX());
		slope2 = -1 / slope1;

		// if the line is horizontal
		if (slope1 == 0) {
			x = spot.getX();
			y = _start.getY();
			// if the line is vertical
		} else if (slope2 == 0) {
			x = _start.getX();
			y = spot.getY();
			// otherwise, the line is sloped
		} else {
			x = (-1 * slope2 * spot.getX() + spot.getY() - _start.getY() + slope1
					* _start.getX())
					/ (slope1 - slope2);
			y = slope1 * (x - _start.getX()) + _start.getY();
		}

		// position spot on the line
		spot.setPosition((int) x, (int) y);
		_virtualSpot = spot;
	}

	public void showVirtualSpot(int mouseX, int mouseY) {
		Item spot = new Dot(mouseX, mouseY, -1);
		spot.setThickness(Math.max(this.getThickness(), 5));
		if (this.getColor() != Color.RED)
			spot.setColor(Color.RED);
		else
			spot.setColor(Color.BLUE);

		// calculate nearest point on line from spot
		double slope1;
		double slope2;
		double x, y;

		slope1 = (_start.getY() - _end.getY() * 1.0)
				/ (_start.getX() - _end.getX());
		slope2 = -1 / slope1;

		// if the line is horizontal
		if (slope1 == 0) {
			x = spot.getX();
			y = _start.getY();
			// if the line is vertical
		} else if (slope2 == 0) {
			x = _start.getX();
			y = spot.getY();
			// otherwise, the line is sloped
		} else {
			x = (-1 * slope2 * spot.getX() + spot.getY() - _start.getY() + slope1
					* _start.getX())
					/ (slope1 - slope2);
			y = slope1 * (x - _start.getX()) + _start.getY();
		}

		// position spot on the line
		spot.setPosition((int) x, (int) y);
		_virtualSpot = spot;
	}

	public Item forceMerge(Item spot, int mouseX, int mouseY) {

		// spot.removeAllConstraints();
		// _end.removeAllConstraints();
		// _start.removeAllConstraints();

		// Keep constraints
		for (Constraint c : _end.getConstraints()) {
			if (c.getOppositeEnd(_end).equals(_start)) {
				c.replaceEnd(_start, spot);
				new Constraint(spot, _start, getParentOrCurrentFrame()
						.getNextItemID(), c.getType());
				break;
			}
		}

		// calculate nearest point on line from spot
		double slope1;
		double slope2;
		double x, y;

		slope1 = (_start.getY() - _end.getY() * 1.0)
				/ (_start.getX() - _end.getX());
		slope2 = -1 / slope1;

		// if the line is horizontal
		if (slope1 == 0) {
			x = spot.getX();
			y = _start.getY();
			// if the line is vertical
		} else if (slope2 == 0) {
			x = _start.getX();
			y = spot.getY();
			// otherwise, the line is sloped
		} else {
			x = (-1 * slope2 * spot.getX() + spot.getY() - _start.getY() + slope1
					* _start.getX())
					/ (slope1 - slope2);
			y = slope1 * (x - _start.getX()) + _start.getY();
		}

		// position spot on the line
		spot.setPosition((int) x, (int) y);

		Line temp = copy();

		Frame currentFrame = DisplayIO.getCurrentFrame();
		temp.setID(currentFrame.getNextItemID());
		temp.setEndItem(_end);
		temp.setStartItem(spot);
		currentFrame.addItem(temp);

		setEndItem(spot);

		// if(_arrowEnd == spot)
		// setArrow(-1,-1);

		return spot;
	}

	@Override
	public Item merge(Item merger, int mouseX, int mouseY) {
		if (!(merger.isLineEnd()))
			return merger;

		Item spot = merger;

		// dots with multiple lines cannot be attached, nor can dots with no
		// lines
		if (spot.getLines().size() != 1)
			return merger;

		// lines that are in 'connected' mode, also cannot be attached
		if (spot.getLines().get(0).getOppositeEnd(spot).isFloating())
			return merger;

		return forceMerge(merger, mouseX, mouseY);
	}

	@Override
	public Collection<Item> getConnected() {
		return getAllConnected();
	}

	@Override
	public void addAllConnected(Collection<Item> connected) {
		if (!connected.contains(this))
			connected.add(this);

		if (!connected.contains(_end))
			_end.addAllConnected(connected);

		if (!connected.contains(_start))
			_start.addAllConnected(connected);
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof Line))
			return false;

		return ((Item) o).getID() == getID();
	}

	@Override
	public void setActions(List<String> actions) {
	}

	@Override
	public List<String> getAction() {
		return null;
	}

	public String getLineEnds() {
		return _start.getID() + " " + _end.getID();
	}

	public int getLineType() {
		return 1;
	}

	@Override
	public boolean isFloating() {
		return _start.isFloating() && _end.isFloating();
	}

	@Override
	protected void updatePolygon() {
		if (_isCircle) {

		}

		_poly = new Polygon();

		Rectangle one;
		Rectangle two;

		/**
		 * TODO: Fake dots seems inefficient, Fix it.
		 */

		// line goes from left->right, one->two
		if (_start.getX() <= _end.getX()) {
			/*
			 * one = _start.getPolygon().getBounds(); two =
			 * _end.getPolygon().getBounds();
			 */

			Item s = new Dot(_start.getX(), _start.getY(), -1);
			Item e = new Dot(_end.getX(), _end.getY(), -1);

			one = s.getPolygon().getBounds();
			two = e.getPolygon().getBounds();
		} else {
			Item s = new Dot(_start.getX(), _start.getY(), -1);
			Item e = new Dot(_end.getX(), _end.getY(), -1);

			one = e.getPolygon().getBounds();
			two = s.getPolygon().getBounds();
			/*
			 * one = _end.getPolygon().getBounds(); two =
			 * _start.getPolygon().getBounds();
			 */
		}

		// if one is above two
		if (one.getY() < two.getY()) {
			_poly.addPoint((int) one.getMaxX(), (int) one.getMinY());
			_poly.addPoint((int) two.getMaxX(), (int) two.getMinY());
			_poly.addPoint((int) two.getMinX(), (int) two.getMaxY());
			_poly.addPoint((int) one.getMinX(), (int) one.getMaxY());
			// if one is below two
		} else {
			_poly.addPoint((int) one.getMinX(), (int) one.getMinY());
			_poly.addPoint((int) two.getMinX(), (int) two.getMinY());
			_poly.addPoint((int) two.getMaxX(), (int) two.getMaxY());
			_poly.addPoint((int) one.getMaxX(), (int) one.getMaxY());
		}
	}

	public void setFilledHighlight(boolean value) {
		_start.setFilledHighlight(value);
		_end.setFilledHighlight(value);
	}

	@Override
	public void delete() {
		super.delete();

		_start.removeLine(this);
		_end.removeLine(this);

		if (_start.getLines().size() == 0)
			_start.delete();
		if (_end.getLines().size() == 0)
			_end.delete();
	}

	@Override
	public void anchor() {
		Frame current = getParentOrCurrentFrame();
		// Check if either end is off the current frame
		Frame parentEnd = getEndItem().getParent();
		if (parentEnd != null && parentEnd != current) {
			this.setParent(parentEnd);
		} else {
			Frame parentStart = getStartItem().getParent();
			if (parentStart != null && parentStart != current)
				this.setParent(parentStart);
		}

		super.anchor();
		fixArrowheadLength();
	}

	/**
	 * Removes constraints from the end points of this line.
	 */
	@Override
	public void removeAllConstraints() {
		// Find all constraints that include both the start and end.
		for (Constraint c : _start.getConstraints()) {
			if (c.contains(_end)) {
				_start.removeConstraint(c);
				_end.removeConstraint(c);
			}
		}
	}
}
