package org.expeditee.items;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextHitInfo;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.text.AttributedString;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.FrameGraphics;
import org.expeditee.gui.FrameKeyboardActions;
import org.expeditee.gui.FrameMouseActions;

/**
 * Represents text displayed on the screen, which may include multiple lines.
 * All standard text properties can be set (font, style, size).
 * 
 * @author jdm18
 * 
 */
public class Text extends Item {

	private String[] _processedText = null;

	public String[] getProcessedText() {
		return _processedText;
	}

	public void setProcessedText(String[] tokens) {
		_processedText = tokens;
	}

	public static final String FRAME_NAME_SEPARATOR = " on frame ";

	/**
	 * The default font used to display text items if no font is specified.
	 */
	public static final String DEFAULT_FONT = "Serif-Plain-18";

	public static final Color DEFAULT_COLOR = Color.BLACK;

	public static final int MINIMUM_RANGED_CHARS = 2;

	public static final int UP = 0;

	public static final int DOWN = 1;

	public static final int LEFT = 2;

	public static final int RIGHT = 3;

	public static final int HOME = 4;

	public static final int END = 5;

	public static final int PAGE_DOWN = 6;

	public static final int PAGE_UP = 7;

	private int _maxWidth = -1;

	private int _justification = Item.JUSTIFICATION_NONE;

	private int _spacing = -1;

	private int _word_spacing = -1;

	private int _initial_spacing = -1;

	private int _letter_spacing = -1;

	// used during ranging out
	private int _selectionStart = -1;

	private int _selectionEnd = -1;

	// text is broken up into lines
	private StringBuffer _text = new StringBuffer();

	private List<TextLayout> _textLayouts = new LinkedList<TextLayout>();

	private List<Integer> _lineOffsets = new LinkedList<Integer>();

	private LineBreakMeasurer _lineBreaker = null;

	// The font to display this text in
	private Font _font;

	// the rectangle surrounding this text item
	private Polygon _poly = null;

	/**
	 * Creates a new Text Item with the given ID and text.
	 * 
	 * @param id
	 *            The id of this item
	 * @param text
	 *            The text to use in this item
	 */
	public Text(int id, String text) {
		super();
		_text.append(text);
		rebuild(false);

		setID(id);
	}

	/**
	 * Creates a new Text Item with the given ID
	 * 
	 * @param id
	 *            The ID to of this item
	 */
	public Text(int id) {
		super();
		setID(id);
	}

	/**
	 * Sets the maximum width of this Text item when justifcation is used.
	 * passing in 0 or -1 means there is no maximum width
	 * 
	 * @param width
	 *            The maximum width of this item when justification is applied
	 *            to it.
	 */
	@Override
	public void setWidth(int width) {
		// 0 is the default
		if (width == 0)
			width = -1;

		_maxWidth = width;
	}

	/**
	 * Returns the maximum width of this Text item when justifcation is used.
	 * Note the returned value may be -1, which indicates that there is no
	 * maximum size set
	 * 
	 * @return The maximum width of this Text item when justification is used
	 */
	@Override
	public int getWidth() {
		return _maxWidth;
	}

	/**
	 * Sets the justification of this Text item. The given integer should
	 * correspond to one of the JUSTIFICATION constants defined in Item
	 * 
	 * @param just
	 *            The justification to apply to this Text item
	 */
	public void setJustification(int just) {
		_justification = just;
		rejustify();
	}

	/**
	 * Returns the current justification of this Text item, the default value is
	 * Item.JUSTIFICATION_NONE
	 * 
	 * @return The justification of this Text item
	 */
	public int getJustification() {
		return _justification;
	}

	private int getJustOffset(TextLayout layout) {
		if (getWidth() < 0)
			return 0;

		if (getJustification() == Item.JUSTIFICATION_CENTER)
			return (int) ((getWidth() - layout.getAdvance()) / 2);
		else if (getJustification() == Item.JUSTIFICATION_RIGHT)
			return (int) (getWidth() - layout.getAdvance());

		return 0;
	}

	/**
	 * Sets the text displayed on the screen to the given String
	 * 
	 * @param text
	 *            The String to display on the screen when drawing this Item.
	 */
	public void setText(String text) {
		_text = new StringBuffer(text);
		rebuild(false);
	}

	public void setTextList(List<String> text) {
		if (text == null || text.size() <= 0)
			return;

		setText(text.get(0));
		for (int i = 1; i < text.size(); i++) {
			appendText("\n");
			appendText(text.get(i));
		}
		rebuild(false);
	}

	public void setAttributeValue(String value) {

	}

	/**
	 * Inserts the given String at the start of the first line of this Text
	 * Item.
	 * 
	 * @param text
	 *            The String to insert.
	 */
	public void prependText(String text) {
		_text.insert(0, text);
		rebuild(false);
	}

	/**
	 * If the first line of text starts with the given String, then it is
	 * removed otherwise no action is taken.
	 * 
	 * @param text
	 *            The String to remove from the first line of Text
	 */
	public void removeText(String text) {
		if (_text.length() > 0 && _text.indexOf(text) == 0)
			_text.delete(0, text.length());

	}

	public void removeEndText(String textToRemove) {
		int length = _text.length();
		if (length > 0) {
			int pos = _text.indexOf(textToRemove);
			int textToRemoveLength = textToRemove.length();
			if (pos + textToRemoveLength == length) {
				_text.delete(pos, length);
			}
		}
	}

	/**
	 * Appends the given String to any text already present in this Item
	 * 
	 * @param text
	 *            The String to append.
	 */
	public void appendText(String text) {
		_text.append(text);
		rebuild(false);
	}

	public void appendLine(String text) {
		if (text == null)
			text = "";

		if (_text.length() > 0)
			_text.append('\n');

		_text.append(text);
		rebuild(false);
	}

	/**
	 * Tests if the first line of this Text starts with the given String.
	 * 
	 * @param text
	 *            The prefix to check for
	 * @return True if the first line starts with the given String, False
	 *         otherwise.
	 */
	public boolean startsWith(String text) {
		return startsWith(text, true);
	}

	public boolean startsWith(String text, boolean ignoreCase) {
		if (text == null || _text == null || _text.length() < 1)
			return false;

		if (ignoreCase)
			return _text.toString().toLowerCase()
					.startsWith(text.toLowerCase());
		else
			return _text.indexOf(text) == 0;
	}

	/**
	 * Inserts a character into the Text of this Item.
	 * 
	 * @param ch
	 *            The character insert.
	 * @param mouseX
	 *            The X position to insert the Strings at.
	 * @param mouseY
	 *            The Y position to insert the Strings at.
	 */
	public Point insertChar(char ch, int mouseX, int mouseY) {
		if (ch != '\t')
			return insertText("" + ch, mouseX, mouseY);

		return insertText(" " + ch, mouseX, mouseY);
	}

	/**
	 * @param index
	 * @return
	 */
	private char getNextBullet(char bullet) {
		switch (bullet) {
		case '*':
			return '>';
		case '>':
			return '+';
		case '+':
			return 'o';
		case 'o':
			return '*';
		}
		return bullet;
	}

	private char getPreviousBullet(char bullet) {
		switch (bullet) {
		case '*':
			return 'o';
		case '>':
			return '*';
		case '+':
			return '>';
		case 'o':
			return '+';
		}
		return bullet;
	}

	public Point getEndPosition(int mouseY) {
		return getEdgePosition(getLinePosition(mouseY), false);
	}

	public Point getStartPosition(int mouseY) {
		return getEdgePosition(getLinePosition(mouseY), true);
	}

	public Point getEndParagraphPosition() {
		return getEdgePosition(_textLayouts.size() - 1, false);
	}
	
	public Point getStartParagraphPosition() {
		return getEdgePosition(0, true);
	}

	private Point getEdgePosition(int line, boolean start) {
		// if there is no text yet, or the line is invalid
		if (_text == null || _text.length() == 0 || line < 0
				|| line > _textLayouts.size() - 1)
			return new Point(getX(), getY());

		TextLayout last = _textLayouts.get(line);
		TextHitInfo hit;
		if (start)
			hit = last.getNextLeftHit(1);
		else
			hit = last.getNextRightHit(last.getCharacterCount() - 1);

		// move the cursor to the new location
		float[] caret = last.getCaretInfo(hit);
		int y = getLineDrop(last) * line;

		int x = ((int) (getX() + caret[0]) + getJustOffset(last));
		x = Math
				.min(
						x,
						(getX() - Item.MARGIN_RIGHT - (2 * getGravity()) + getBoundsWidth()));
		return new Point(x, getY() + /* DisplayIO.getOffset() + */y
				+ (int) caret[1]);
	}

	public void setSelectionStart(int mouseX, int mouseY) {
		// determine what line is being pointed to
		int line = getLinePosition(mouseY);

		// get the character being pointed to
		TextHitInfo hit = getCharPosition(line, mouseX);
		_selectionStart = hit.getInsertionIndex() + _lineOffsets.get(line);
	}

	public void setSelectionEnd(int mouseX, int mouseY) {
		// determine what line is being pointed to
		int line = getLinePosition(mouseY);

		// get the character being pointed to
		TextHitInfo hit = getCharPosition(line, mouseX);
		_selectionEnd = hit.getInsertionIndex() + _lineOffsets.get(line);
	}

	public void clearSelection() {
		_selectionStart = -1;
		_selectionEnd = -1;
	}

	public void clearSelectionEnd() {
		_selectionEnd = -1;
	}

	public String copySelectedText() {
		if (_selectionStart < 0 || _selectionEnd < 0)
			return null;
		else if (_selectionEnd > _text.length())
			_selectionEnd = _text.length();

		return _text.substring(Math.min(_selectionStart, _selectionEnd), Math
				.max(_selectionStart, _selectionEnd));
	}

	public String cutSelectedText() {
		return replaceSelectedText("");
	}

	public String replaceSelectedText(String newText) {
		if (_selectionStart < 0 || _selectionEnd < 0)
			return null;
		else if (_selectionEnd > _text.length())
			_selectionEnd = _text.length();

		int left = Math.min(_selectionStart, _selectionEnd);
		int right = Math.max(_selectionStart, _selectionEnd);

		String s = _text.substring(left, right);

		_text.delete(left, right);
		_text.insert(left, newText);
		rebuild(true);

		return s;
	}

	public int getSelectedSize() {
		if (_selectionEnd < 0 || _selectionStart < 0)
			return 0;

		// System.out.println(_selectionStart + ":" + _selectionEnd);

		return Math.abs(_selectionEnd - _selectionStart);
	}

	/**
	 * Inserts the given String into the Text at the position given by the
	 * mouseX and mouseY coordinates
	 * 
	 * @param text
	 *            The String to insert into this Text.
	 * @param mouseX
	 *            The X position to insert the String
	 * @param mouseY
	 *            The Y position to insert the String
	 * @return The new location that the mouse cursor should be moved to
	 */
	public java.awt.Point insertText(String text, int mouseX, int mouseY) {
		TextHitInfo hit;
		TextLayout current = null;
		int line;

		// check for empty string
		if (text == null || text.length() == 0)
			return new Point(mouseX, mouseY);

		// if there is no text yet
		if (_text == null || _text.length() == 0) {
			_text = new StringBuffer().append(text);
			// create the linebreaker and layouts
			rebuild(true);
			assert (_textLayouts.size() == 1);
			current = _textLayouts.get(0);
			hit = current.getNextRightHit(0);
			line = 0;

			// otherwise, we are inserting text
		} else {
			_processedText = null;
			// determine what line is being pointed to
			line = getLinePosition(mouseY);

			// get the character being pointed to
			hit = getCharPosition(line, mouseX);

			int pos = hit.getInsertionIndex() + _lineOffsets.get(line);

			if (line > 0 && hit.getInsertionIndex() == 0)
				pos++;

			// if this is a backspace key
			if (text.charAt(0) == KeyEvent.VK_BACK_SPACE) {
				if (pos > 0) {
					deleteChar(pos - 1, false);

					pos--;
				}
				// if this is a delete key
			} else if (text.charAt(0) == KeyEvent.VK_DELETE) {
				if (pos < _text.length()) {
					deleteChar(pos, false);
				}
				// this is a tab
			} else if (text.charAt(0) == KeyEvent.VK_TAB) {
				// Text length greater than 1 signals a backwards tab
				if (text.length() > 1) {
					// Find the first non space char to see if its a bullet
					int index = 0;
					for (index = 0; index < _text.length(); index++) {
						if (!Character.isSpaceChar(_text.charAt(index)))
							break;
					}
					// Check if there is a space after the bullet
					if (index < _text.length() - 1
							&& _text.charAt(index + 1) == ' ') {
						// Change the bullet
						_text.setCharAt(index, getPreviousBullet(_text
								.charAt(index)));
					}
					// Remove the spacing at the start
					for (int i = 0; i < TAB_STRING.length(); i++) {
						if (_text.length() > 0
								&& Character.isSpaceChar(_text.charAt(0))) {
							deleteChar(0, false);
							pos--;
						} else
							break;
					}
					_lineBreaker = null;
				} else {
					// / Find the first non space char to see if its a bullet
					int index = 0;
					for (index = 0; index < _text.length(); index++) {
						if (!Character.isSpaceChar(_text.charAt(index)))
							break;
					}
					// Check if there is a space after the bullet
					if (index < _text.length() - 1
							&& _text.charAt(index + 1) == ' ') {
						char nextBullet = getNextBullet(_text.charAt(index));
						// Change the bullet
						_text.setCharAt(index, nextBullet);
					}
					// Insert the spacing at the start
					insertString(TAB_STRING, 0);
					pos += TAB_STRING.length();
				}
				// this is a normal insert
			} else {
				insertString(text, pos);
				pos += text.length();
			}

			int newLine = line;

			// if a rebuild is required
			if (_lineBreaker == null)
				rebuild(true);
			else
				rejustify();

			// determine the new position the cursor should have
			for (int i = 1; i < _lineOffsets.size(); i++) {
				if (_lineOffsets.get(i) >= pos) {
					newLine = i - 1;
					break;
				}
			}

			current = _textLayouts.get(newLine);
			pos -= _lineOffsets.get(newLine);

			if (newLine == line) {
				if (pos > 0)
					hit = current.getNextRightHit(pos - 1);
				else
					hit = current.getNextLeftHit(1);
			} else if (newLine < line) {
				hit = current.getNextRightHit(pos - 1);
			} else {
				hit = current.getNextRightHit(pos - 1);
			}

			line = newLine;
		}

		// move the cursor to the new location
		float[] caret = current.getCaretInfo(hit);
		int y = getLineDrop(current) * line;

		int x = ((int) (getX() + caret[0]) + getJustOffset(current));
		x = Math
				.min(
						x,
						(getX() - Item.MARGIN_RIGHT - (2 * getGravity()) + getBoundsWidth()));
		return new Point(x, getY() + y + (int) caret[1]);
	}

	public Point moveCursor(int direction, int mouseX, int mouseY) {
		// check for home or end keys
		if (direction == HOME)
			return getStartParagraphPosition();

		if (direction == END)
			return getEndParagraphPosition();

		TextHitInfo hit;
		TextLayout current;
		int line;

		// if there is no text yet
		if (_text == null || _text.length() == 0) {
			return new Point(mouseX, mouseY);
			// otherwise, move the cursor
		} else {
			// determine the line of text to check
			line = getLinePosition(mouseY);
			if (line < 0)
				line = _textLayouts.size() - 1;

			// if the cursor is moving up or down, change the line
			if (direction == UP)
				line = Math.max(line - 1, 0);
			else if (direction == DOWN)
				line = Math.min(line + 1, _textLayouts.size() - 1);

			hit = getCharPosition(line, mouseX);

			if (direction == LEFT) {
				if (hit.getInsertionIndex() > 0) {
					hit = _textLayouts.get(line).getNextLeftHit(hit);
					//This takes care of the case where the user has put a hard line break in
					if (line > 0
							&& _text.charAt(hit.getInsertionIndex()
									+ _lineOffsets.get(line)) == '\n') {
						line--;
						hit = _textLayouts.get(line).getNextRightHit(
								_textLayouts.get(line).getCharacterCount() - 1);
					}
					//This takes care of soft line breaks.
				} else if (line > 0) {
					line--;
					hit = _textLayouts.get(line).getNextRightHit(
							_textLayouts.get(line).getCharacterCount() - 1);
					while (hit.getCharIndex() > 0
							&& _text.charAt(_lineOffsets.get(line)
									+ hit.getCharIndex()- 1) == ' '){
						hit = _textLayouts.get(line).getNextLeftHit(hit);
					}
				}
			} else if (direction == RIGHT) {
				if (hit.getInsertionIndex() < _textLayouts.get(line)
						.getCharacterCount()) {
					hit = _textLayouts.get(line).getNextRightHit(hit);
					while (hit.getCharIndex() > 0
							&& hit.getCharIndex() < _textLayouts.get(line)
									.getCharacterCount()
							&& _text.charAt(_lineOffsets.get(line)
									+ hit.getCharIndex() - 1) == '\t')
						hit = _textLayouts.get(line).getNextRightHit(hit);
				} else if (line < _textLayouts.size() - 1) {
					line++;
					hit = _textLayouts.get(line).getNextLeftHit(1);
				}
			}

			current = _textLayouts.get(line);
		}

		// move the cursor to the new location
		float[] caret = current.getCaretInfo(hit);
		int y = getLineDrop(current) * line;

		return new Point((int) (getX() + caret[0]) + getJustOffset(current),
				getY() + y + (int) caret[1]);
	}

	/**
	 * Iterates through the given line string and returns the position of the
	 * character being pointed at by the mouse.
	 * 
	 * @param line
	 *            The index of the _text array of the String to be searched.
	 * @param mouseX
	 *            The X coordinate of the mouse
	 * @return The position in the string of the character being pointed at.
	 */
	private TextHitInfo getCharPosition(int line, int mouseX) {
		if (line < 0 || line >= _textLayouts.size())
			return null;

		TextLayout layout = _textLayouts.get(line);
		mouseX += getOffset().x;
		mouseX -= getJustOffset(layout);

		return layout.hitTestChar(mouseX - getX(), 0);
	}

	private int getLinePosition(int mouseY) {
		mouseY += getOffset().y;

		int y = getY();

		for (TextLayout text : _textLayouts) {
			// calculate X to ensure it is in the shape
			Rectangle bounds = text.getLogicalHighlightShape(0,
					text.getCharacterCount()).getBounds();

			if (bounds.getWidth() < 1)
				bounds = new Rectangle((int) bounds.getMinX(), (int) bounds
						.getMinY(), 10, (int) bounds.getHeight());

			double x = bounds.getCenterX();

			// if(text.getLogicalHighlightShape(0,
			// text.getCharacterCount()).contains(text.getBounds().getMinX()
			// + 10, mouseY - getY() - (y - getY())))
			if (bounds.contains(x, mouseY - getY() - (y - getY())))
				return _textLayouts.indexOf(text);

			// check if the cursor is between lines
			if (mouseY - getY() - (y - getY()) < bounds.getMinY())
				return Math.max(0, _textLayouts.indexOf(text) - 1);

			y += getLineDrop(text);
		}

		return _textLayouts.size() - 1;
	}

	/**
	 * Sets the Font that this text will be displayed with on the screen.
	 * 
	 * @param font
	 *            The Font to display the Text of this Item in.
	 */
	public void setFont(Font font) {
		// all decoding occurs in the Utils class
		_font = font;

		// rejustify();
		rebuild(false);
	}

	/**
	 * Returns the Font that this Text is currently using when painting to the
	 * screen
	 * 
	 * @return The Font used to display this Text on the screen.
	 */
	public Font getFont() {
		return _font;
	}

	public Font getPaintFont() {
		if (getFont() == null)
			return Font.decode(DEFAULT_FONT);

		return getFont();
	}

	public String getFamily() {
		return getPaintFont().getFamily();
	}

	public void setFamily(String newFamily) {
		String toDecode = newFamily + "-" + getFontStyle() + "-" + getSize();
		setFont(Font.decode(toDecode));
	}

	public String getFontStyle() {
		Font f = getPaintFont();
		String s = "";

		if (f.isPlain())
			s += "Plain";

		if (f.isBold())
			s += "Bold";

		if (f.isItalic())
			s += "Italic";

		return s;
	}

	public static final String MONOSPACED_FONT = "monospaced";

	public static final String[] FONT_WHEEL = { "sansserif", "monospaced",
			"serif", "dialog", "dialoginput" };

	public static final char[] FONT_CHARS = { 's', 'm', 't', 'd', 'i' };

	private static final int NEARBY_GRAVITY = 2;

	private static final int MINIMUM_FONT_SIZE = 8;

	public void toggleFontFamily() {
		String fontFamily = getFamily().toLowerCase();
		// set it to the first font by default
		setFamily(FONT_WHEEL[0]);

		for (int i = 0; i < FONT_WHEEL.length - 3; i++) {
			if (fontFamily.equals(FONT_WHEEL[i])) {
				setFamily(FONT_WHEEL[i + 1]);
				break;
			}
		}
	}

	public void toggleFontStyle() {
		Font currentFont = getPaintFont();
		if (currentFont.isPlain())
			setFont(currentFont.deriveFont(Font.BOLD));
		else if (currentFont.isBold() && currentFont.isItalic())
			setFont(currentFont.deriveFont(Font.PLAIN));
		else if (currentFont.isBold())
			setFont(currentFont.deriveFont(Font.ITALIC));
		else
			setFont(currentFont.deriveFont(Font.ITALIC + Font.BOLD));
	}

	public void setFontStyle(String newFace) {
		if (newFace == null || newFace.trim().length() == 0) {
			setFont(getPaintFont().deriveFont(Font.PLAIN));
			return;
		}

		newFace = newFace.toLowerCase().trim();

		if (newFace.equals("plain")) {
			setFont(getPaintFont().deriveFont(Font.PLAIN));
		} else if (newFace.equals("bold")) {
			setFont(getPaintFont().deriveFont(Font.BOLD));
		} else if (newFace.equals("italic")) {
			setFont(getPaintFont().deriveFont(Font.ITALIC));
		} else if (newFace.contains("bold") && newFace.contains("italic")) {
			setFont(getPaintFont().deriveFont(Font.BOLD + Font.ITALIC));
		}

	}

	/**
	 * Returns a String array of this Text object's text, split up into separate
	 * lines.
	 * 
	 * @return The String array with one element per line of text in this Item.
	 */
	public List<String> getTextList() {
		if (_text == null)
			return null;
		try {
			List<String> list = new LinkedList<String>();

			int last = 0;
			for (int offset : _lineOffsets) {
				if (offset != last) {
					list
							.add(_text.substring(last, offset).replaceAll("\n",
									""));
				}
				last = offset;
			}

			return list;
		} catch (Exception e) {
			System.out.println(e.getMessage());
			return null;
		}
	}

	public String getText() {
		return _text.toString();
	}

	/**
	 * Returns the first line of text in this Text Item
	 * 
	 * @return The first line of Text
	 */
	public String getFirstLine() {
		if (_text == null || _text.length() == 0)
			return null;

		if (_text.indexOf("\n") < 0)
			return _text.toString();

		return _text.substring(0, _text.indexOf("\n"));
	}

	/**
	 * Sets the inter-line spacing (in pixels) of this text.
	 * 
	 * @param spacing
	 *            The number of pixels to allow between each line
	 */
	public void setSpacing(int spacing) {
		_spacing = spacing;
		updatePolygon();
	}

	/**
	 * Returns the inter-line spacing (in pixels) of this Text.
	 * 
	 * @return The spacing (inter-line) in pixels of this Text.
	 */
	public int getSpacing() {
		return _spacing;
	}

	private int getLineDrop(TextLayout layout) {
		if (getSpacing() < 0)
			return (int) (layout.getAscent() + layout.getDescent() + layout
					.getLeading());

		return (int) (layout.getAscent() + layout.getDescent() + getSpacing());
	}

	public void setWordSpacing(int spacing) {
		_word_spacing = spacing;
	}

	public int getWordSpacing() {
		return _word_spacing;
	}

	public void setLetterSpacing(int spacing) {
		_letter_spacing = spacing;
	}

	public int getLetterSpacing() {
		return _letter_spacing;
	}

	public void setInitialSpacing(int spacing) {
		_initial_spacing = spacing;
	}

	public int getInitialSpacing() {
		return _initial_spacing;
	}

	@Override
	public boolean intersects(Polygon p) {
		if (super.intersects(p)) {
			int textY = getY();

			for (TextLayout text : _textLayouts) {
				// check left and right of each box
				Rectangle textOutline = text.getLogicalHighlightShape(0,
						text.getCharacterCount()).getBounds();
				textOutline.width += 2;
				textOutline.height += 2;

				textOutline.translate(getX() - 1, textY - 1);
				if (p.intersects(textOutline))
					return true;
				textY += getLineDrop(text);
			}
		}
		return false;
	}

	@Override
	public boolean contains(int mouseX, int mouseY) {
		return contains(mouseX, mouseY, getGravity() * NEARBY_GRAVITY);
	}

	public boolean contains(int mouseX, int mouseY, int gravity) {
		mouseX += getOffset().x;
		mouseY += getOffset().y;

		int textY = getY();
		int textX = getX();

		Rectangle outline = getPolygon().getBounds();

		// Check if its outside the top and left and bottom bounds
		if (outline.x - mouseX > gravity || outline.y - mouseY > gravity
				|| mouseY - (outline.y + outline.height) > gravity
				|| mouseX - (outline.x + outline.width) > gravity) {
			return false;
		}

		for (TextLayout text : _textLayouts) {
			// check left and right of each box
			Rectangle textOutline = text.getLogicalHighlightShape(0,
					text.getCharacterCount()).getBounds();

			// check if the cursor is within the top, bottom and within the
			// gravity of right
			if (mouseY - textY > textOutline.y
					&& mouseY - textY < textOutline.y + textOutline.height
					&& mouseX - textX < textOutline.width + gravity
							+ Item.MARGIN_RIGHT)
				return true;
			textY += getLineDrop(text);
		}

		return false;
	}

	/**
	 * Updates the Polygon (rectangle) that surrounds this Text on the screen.
	 */
	protected void updatePolygon() {
		// if there is no text, there is nothing to do
		if (_text == null)
			return;

		_poly = new Polygon();

		if (_textLayouts.size() < 1)
			return;

		int minX = Integer.MAX_VALUE;
		int maxX = Integer.MIN_VALUE;

		int minY = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;

		float y = -1;

		synchronized (_textLayouts) {
			for (TextLayout layout : _textLayouts) {
				Rectangle2D bounds = layout.getLogicalHighlightShape(0,
						layout.getCharacterCount()).getBounds2D();

				if (y < 0)
					y = 0;
				else
					y += getLineDrop(layout);

				maxX = Math.max(maxX, (int) bounds.getMaxX());
				minX = Math.min(minX, (int) bounds.getMinX());
				maxY = Math.max(maxY, (int) (bounds.getMaxY() + y));
				minY = Math.min(minY, (int) (bounds.getMinY() + y));
			}
		}

		minX -= getLeftMargin();
		maxX += Item.MARGIN_RIGHT;

		_poly.addPoint(minX - getGravity(), minY - getGravity());
		_poly.addPoint(maxX + getGravity(), minY - getGravity());
		_poly.addPoint(maxX + getGravity(), maxY + getGravity());
		_poly.addPoint(minX - getGravity(), maxY + getGravity());
	}

	// TODO it seems like this method has some exponencial processing which
	// makes items copy really slowly when there are lots of lines of text!
	// This needs to be fixed!!
	private void rebuild(boolean limitWidth) {
		// if there is no text, there is nothing to do
		if (_text == null || _text.length() == 0)
			return;

		AttributedString paragraphText = new AttributedString(_text.toString());
		paragraphText.addAttribute(TextAttribute.FONT, getPaintFont());
		_lineBreaker = new LineBreakMeasurer(paragraphText.getIterator(),
				new FontRenderContext(null, true, true));

		float width = Float.MAX_VALUE;

		if (limitWidth) {
			if (getWidth() > 0)
				width = getWidth();
			else if (getMaxSize() != null)
				width = Math.max(50, getMaxSize().width - getX()
						- Item.MARGIN_RIGHT);
		}

		_textLayouts.clear();
		_lineOffsets.clear();
		// the first line always has a 0 offset
		_lineOffsets.add(0);

		TextLayout layout;
		_lineBreaker.setPosition(0);

		// --- Get the output of the LineBreakMeasurer and store it in a
		while ((layout = _lineBreaker.nextLayout(width)) != null) {
			// for some reason lineBreaker will not break on newline
			// characters so they have to be check manually
			int start = _lineOffsets.get(_lineOffsets.size() - 1);

			// check through the current line for newline characters
			for (int i = start + 1; i < _text.length(); i++) {
				if (_text.charAt(i) == '\n') {// || c == '\t'){
					_lineBreaker.setPosition(start);
					layout = _lineBreaker.nextLayout(width, i, false);
					break;
				}
			}

			_lineOffsets.add(_lineBreaker.getPosition());

			if (getWidth() > 0 && getJustification() == Item.JUSTIFICATION_FULL
					&& _lineBreaker.getPosition() < _text.length())
				layout = layout.getJustifiedLayout(width);

			_textLayouts.add(layout);
		}

		updatePolygon();
	}

	private void rejustify() {
		// if there is no text, there is nothing to do
		if (_text == null || _text.length() == 0)
			return;

		// only recreate linebreaker if necessary
		if (_lineBreaker == null) {
			rebuild(true);
			return;

			/*
			 * AttributedString paragraphText = new
			 * AttributedString(_text.toString());
			 * paragraphText.addAttribute(TextAttribute.FONT, getPaintFont());
			 * _lineBreaker = new LineBreakMeasurer(paragraphText.getIterator(),
			 * new FontRenderContext(null, true, true));
			 */
		}

		float width = Float.MAX_VALUE;
		if (getWidth() > 0)
			width = getWidth();
		else if (getMaxSize() != null)
			width = Math.max(50, getMaxSize().width - getX()
					- Item.MARGIN_RIGHT);

		_textLayouts.clear();
		_lineOffsets.clear();
		// the first line always has a 0 offset
		_lineOffsets.add(0);

		TextLayout layout;
		_lineBreaker.setPosition(0);

		// --- Get the output of the LineBreakMeasurer and store it in a
		while ((layout = _lineBreaker.nextLayout(width)) != null) {
			// for some reason lineBreaker will not break on newline
			// characters
			// so they have to be check manually
			int start = _lineOffsets.get(_lineOffsets.size() - 1);

			// check through the current line for newline characters
			for (int i = start + 1; i < _text.length(); i++) {
				if (_text.charAt(i) == '\n') {// || c == '\t'){
					_lineBreaker.setPosition(start);
					layout = _lineBreaker.nextLayout(width, i, false);
					break;
				}
			}

			_lineOffsets.add(_lineBreaker.getPosition());

			if (getWidth() > 0 && getJustification() == Item.JUSTIFICATION_FULL
					&& _lineBreaker.getPosition() < _text.length())
				layout = layout.getJustifiedLayout(width);

			_textLayouts.add(layout);
		}

		updatePolygon();
	}

	private int _alpha = -1;

	public void setAlpha(int alpha) {
		_alpha = alpha;
	}

	private Point getSelectedRange(int line) {
		if (_selectionEnd >= _text.length()) {
			_selectionEnd = _text.length();
		}

		if (_selectionStart < 0)
			_selectionStart = 0;

		if (_selectionStart < 0 || _selectionEnd < 0)
			return null;

		int selectionLeft = Math.min(_selectionStart, _selectionEnd);
		int selectionRight = Math.max(_selectionStart, _selectionEnd);

		// if the selection is after this line, return null
		if (_lineOffsets.get(line) > selectionRight)
			return null;

		// if the selection is before this line, return null
		if (_lineOffsets.get(line) < selectionLeft
				&& _lineOffsets.get(line)
						+ _textLayouts.get(line).getCharacterCount() < selectionLeft)
			return null;

		// Dont highlight a single char
		// if (selectionRight - selectionLeft <= MINIMUM_RANGED_CHARS)
		// return null;

		// the selection occurs on this line, determine where it lies on the
		// line
		int start = Math.max(0, selectionLeft - _lineOffsets.get(line));
		// int end = Math.min(_lineOffsets.get(line) +
		// _textLayouts.get(line).getCharacterCount(), _selectionEnd);
		int end = Math.min(selectionRight - _lineOffsets.get(line),
				+_textLayouts.get(line).getCharacterCount());

		// System.out.println(line + ": " + start + "x" + end + " (" +
		// _selectionStart + "x" + _selectionEnd + ")");
		return new Point(start, end);
	}

	@Override
	public void paint(Graphics2D g) {
		// if there is no text to paint, do nothing.
		if (_text == null || _text.length() == 0 || getMaxSize() == null)// ||
			// _textLayouts.size()
			// < 1)
			return;

		if (_textLayouts.size() < 1) {
			rebuild(true);
			System.out.println("Error: " + _text);
			return;
		}

		// the background is only cleared if required
		if (getBackgroundColor() != null) {
			Color bgc = getBackgroundColor();
			if (_alpha > 0)
				bgc = new Color(bgc.getRed(), bgc.getGreen(), bgc.getBlue(),
						_alpha);
			g.setColor(bgc);
			g.fill(getPolygon());
		}

		float y = getY();
		Color c = getPaintColor();
		if (_alpha > 0)
			c = new Color(c.getRed(), c.getGreen(), c.getBlue(), _alpha);

		g.setColor(c);

		Color selection;

		/*
		 * Color main = getPaintColor(); Color back = getPaintBackgroundColor();
		 * 
		 * if (Math.abs(main.getRed() - back.getRed()) < 10 &&
		 * Math.abs(main.getGreen() - back.getGreen()) < 10 &&
		 * Math.abs(main.getBlue() - back.getBlue()) < 10) { selection = new
		 * Color(Math.abs(255 - main.getRed()), Math .abs(255 -
		 * main.getGreen()), Math.abs(255 - main.getBlue())); } else { selection =
		 * new Color((main.getRed() + (back.getRed() * 2)) / 3, (main.getGreen() +
		 * (back.getGreen() * 2)) / 3, (main .getBlue() + (back.getBlue() * 2)) /
		 * 3); }
		 */
		int green = 160;
		int red = 160;
		int blue = 160;

		if (FrameMouseActions.wasDeleteClicked()) {
			green = 235;
			red = 235;
			blue = 140;
		} else if (FrameMouseActions.getLastMouseButton() == MouseEvent.BUTTON1) {
			red = 255;
		} else if (FrameMouseActions.getLastMouseButton() == MouseEvent.BUTTON2) {
			green = 255;
		} else if (FrameMouseActions.getLastMouseButton() == MouseEvent.BUTTON3) {
			blue = 255;
		}
		selection = new Color(red, green, blue);

		// width -= getX();
		// int line = 0;
		// boolean tab = false;
		synchronized (_textLayouts) {
			for (int i = 0; i < _textLayouts.size(); i++) {
				TextLayout layout = _textLayouts.get(i);

				Point p = getSelectedRange(i);
				if (p != null) {
					AffineTransform at = new AffineTransform();
					AffineTransform orig = g.getTransform();
					at.translate(getX() + getJustOffset(layout), y);
					g.setTransform(at);

					g.setColor(selection);
					g.fill(layout.getLogicalHighlightShape(p.x, p.y));

					g.setTransform(orig);
					g.setColor(c);
				}

				layout.draw(g, getX() + getJustOffset(layout), y);

				/*
				 * AffineTransform at = new AffineTransform(); AffineTransform
				 * orig = g.getTransform(); at.translate(getX() +
				 * getJustOffset(layout), y); g.setTransform(at);
				 * g.draw(layout.getLogicalHighlightShape(0,
				 * layout.getCharacterCount())); g.setTransform(orig); /*
				 * if(_text.charAt(_lineOffsets.get(line) +
				 * (layout.getCharacterCount() - 1)) == '\t'){ tab = true; x =
				 * (int) (getX() + x + 20 + layout.getVisibleAdvance()); }else{
				 */
				y += getLineDrop(layout);
				/*
				 * tab = false; }
				 * 
				 * line++;
				 */
			}
		}

		paintLink(g);

		if (isHighlighted()) {
			g.setColor(getPaintHighlightColor());
			Stroke _highlightStroke = new BasicStroke(
					(float) _highlightThickness, BasicStroke.CAP_BUTT,
					BasicStroke.JOIN_ROUND);
			g.setStroke(_highlightStroke);
			g.drawPolygon(getPolygon());
		}
	}

	/**
	 * Determines if this text has any text in it.
	 * 
	 * @return True if this Item has no text in it, false otherwise.
	 */
	public boolean isEmpty() {
		if (_text == null || _text.length() == 0)
			return true;

		return false;
	}

	@Override
	public Polygon getPolygon() {
		updatePolygon();

		Polygon external = new Polygon(_poly.xpoints, _poly.ypoints,
				_poly.npoints);
		external.translate(getX(), getY());

		return external;
	}

	@Override
	public Text copy() {
		Text copy = new Text(getID());

		// copy standard item values
		Item.DuplicateItem(this, copy);

		// copy values specific to text items
		copy.setSpacing(getSpacing());
		copy.setInitialSpacing(getInitialSpacing());
		copy.setJustification(getJustification());
		copy.setLetterSpacing(getLetterSpacing());
		copy.setWordSpacing(getWordSpacing());
		copy.setWidth(getWidth());
		copy.setFont(getFont());
		copy.setText(_text.toString());

		return copy;
	}

	@Override
	public int getSize() {
		return getPaintFont().getSize();
	}

	/**
	 * Returns the number of characters in this Text, excluding new lines.
	 * 
	 * @return The sum of the length of each line of text
	 */
	public int getLength() {
		return _text.length();
	}

	@Override
	public void setSize(int size) {
		// Dont want to have size set when duplicating a point which has size 0
		if (size < 0)
			return;

		if (size < MINIMUM_FONT_SIZE)
			size = MINIMUM_FONT_SIZE;
		setFont(getPaintFont().deriveFont((float) size));
	}

	@Override
	public void setAnnotation(boolean val) {
		if (val) {
			// if this is already an annotation, do nothing
			if (isAnnotation())
				return;
			if (_text.charAt(0) == '*') {
				deleteChar(0, true);
				if (_text.length() > 0 && _text.charAt(0) == ' ')
					deleteChar(0, true);
			} else {
				insertString("@", 0);
			}
		} else {
			// if this is not an annotation, do nothing
			if (!isAnnotation())
				return;

			if (_text.charAt(0) == '@') {
				_text.setCharAt(0, ' ');
				insertString("*", 0);
			} else {
				deleteChar(0, true);
			}
		}
		rebuild(false);
	}

	/**
	 * 
	 */
	private void insertString(String toInsert, int pos) {
		assert (toInsert.length() > 0);

		_text.insert(pos, toInsert);

		if (toInsert.length() > 1) {
			_lineBreaker = null;
		}

		if (_lineBreaker != null) {
			AttributedString inserting = new AttributedString(_text.toString());
			inserting.addAttribute(TextAttribute.FONT, getPaintFont());
			_lineBreaker.insertChar(inserting.getIterator(), pos);
		}
	}

	private void deleteChar(int pos, boolean replaceWithDot) {
		_text.deleteCharAt(pos);

		if (_text.length() == 0) {
			if (replaceWithDot) {
				// Remove and replace with a dot
				FrameKeyboardActions.replaceText(this);
				DisplayIO.setCursorPosition(this.getPosition());
			}
			return;
		}

		if (_lineBreaker != null) {
			AttributedString inserting = new AttributedString(_text.toString());
			inserting.addAttribute(TextAttribute.FONT, getPaintFont());
			_lineBreaker.deleteChar(inserting.getIterator(), pos);
		}
	}

	@Override
	public boolean isAnnotation() {
		if (_text != null && _text.length() > 0 && _text.charAt(0) == '@')
			return true;

		return false;
	}

	public boolean isSpecialAnnotation() {
		assert _text != null;
		String s = _text.toString().toLowerCase();
		if (s.length() > 0 && s.indexOf("@") == 0) {
			if (s.equals("@old") || s.equals("@ao")
					|| s.equals("@itemtemplate") || s.equals("@parent")
					|| s.equals("@more") || s.equals("@next")
					|| s.equals("@previous") || s.equals("@i")
					|| s.equals("@f"))
				return true;
		}

		return false;
	}

	@Override
	public Item merge(Item merger, int mouseX, int mouseY) {

		if (merger.isLineEnd()) {
			//Merging line ends onto non line end text is a no-op
			if(!isLineEnd())
				return null;
			
			if (merger instanceof Text)
				insertText(((Text) merger).getText(), mouseX, mouseY);

			// Set the position by moving the cursor before calling this
			// method!!

			List<Line> lines = new LinkedList<Line>();
			lines.addAll(merger.getLines());
			for (Line line : lines) {
				line.replaceEnd(merger, this);
			}
			merger.delete();
			this.setOffset(0, 0);
			return null;
		}

		if (!(merger instanceof Text))
			return merger;

		Text merge = (Text) merger;

		// insertText(merge.getText(), mouseX, mouseY);

		// if the item being merged has a link
		if (merge.getLink() != null) {
			// if this item has a link, keep it on the cursor
			if (getLink() != null) {
				merge.setText(merge.getLink());
				merge.setLink(null);
				// return merge;
				// TODO get this to return the merged item and attach it to the
				// cursor only when the user presses the middle button.
			} else
				setLink(merge.getLink());
		}

		return null;
	}

	/**
	 * Resets the position of the item to the default position for a title item.
	 * 
	 */
	public void resetTitlePosition() {
		setPosition(MARGIN_LEFT, MARGIN_LEFT + getBoundsHeight());
	}

	/**
	 * Removes the set of characters up to the first space in this text item.
	 * 
	 * @return the string that was removed.
	 */
	public String stripFirstWord() {
		int firstSpace = _text.toString().indexOf(' ');

		// if there is only one word just make it blank
		if (firstSpace < 0 || firstSpace + 1 >= _text.length()) {
			String text = _text.toString();
			setText("");
			return text;
		}

		String firstWord = _text.toString().substring(0, firstSpace);
		setText(_text.toString().substring(firstSpace).trim());

		return firstWord;
	}

	public String toString() {
		String message = "[" + getFirstLine() + "]" + FRAME_NAME_SEPARATOR;

		if (getParent() != null)
			return message + getParent().getName();
		return message + getDateCreated();
	}

	public Text getTemplateForm() {
		Text template = this.copy();
		template.setID(-1);
		// The template must have text otherwise the bounds height will be
		// zero!!
		// This will stop escape drop down from working if there is no item
		// template
		template.setText("@");
		return template;
	}

	@Override
	public boolean isNear(int x, int y) {
		if (super.isNear(x, y)) {
			// TODO check that it is actually near one of the lines of space
			// return contains(x, y, getGravity() * 2 + NEAR_DISTANCE);
			// at the moment contains ignores gravity when checking the top and
			// bottom of text lines... so the cursor must be between two text
			// lines
			int textY = getY();
			int textX = getX();

			for (TextLayout text : _textLayouts) {
				// check left and right of each box
				Rectangle textOutline = text.getLogicalHighlightShape(0,
						text.getCharacterCount()).getBounds();

				// check if the cursor is within the top, bottom and within the
				// gravity of right
				if (y - textY > textOutline.y - NEAR_DISTANCE
						&& y - textY < textOutline.y + textOutline.height
								+ NEAR_DISTANCE
						&& x - textX < textOutline.width + NEAR_DISTANCE)
					return true;
				textY += getLineDrop(text);
			}
		}
		return false;
	}

	@Override
	public void delete() {
		setVisible(false);
	}

	@Override
	public void anchor() {
		super.anchor();
		// ensure all text items have their selection cleared
		clearSelection();
		setAlpha(0);
	}

	public void resetFrameNamePosition() {
		Dimension maxSize = FrameGraphics.getMaxFrameSize();
		setMaxSize(maxSize);
		if (maxSize != null) {
			setPosition(maxSize.width - getBoundsWidth(), getBoundsHeight());
		}
	}

	@Override
	protected int getLinkYOffset() {
		return Math.round(-(_textLayouts.get(0).getAscent() / 2));
	}

	@Override
	public String getName() {
		return getFirstLine();
	}

	public static final String TAB_STRING = "      ";

	public Point insertTab(char ch, int mouseX, int mouseY) {
		return insertText("" + ch, mouseX, mouseY);
	}

	public Point removeTab(char ch, int mouseX, int mouseY) {
		// Insert a space as a flag that it is a backwards tab
		return insertText(ch + " ", mouseX, mouseY);
	}
}
